/*
 * Firmware framework for USB I/O on PIC 18F2455 (and siblings)
 * Copyright (C) 2005 Alexander Enzmann
 * adapted to MCC18 by Alberto Maccioni on 1/8/09
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 * or see <http://www.gnu.org/licenses/>
 */

#include <p18cxxx.h>
#include <stdio.h>
#include <usart.h>
#include "usb.h"

// Note: there are timing related problems associated with GET_FEATURE
// when run at less than 48 MHz
#define CLK_48MHZ 1

#pragma config PLLDIV=3, CPUDIV=OSC1_PLL2, USBDIV=2, FOSC=HSPLL_HS//, FCMEM=OFF
#pragma config IESO=OFF, PWRT=OFF, BOR=ON_ACTIVE, BORV=3, VREGEN=ON, WDT=OFF
#pragma config MCLRE=OFF, PBADEN=OFF, CCP2MX=OFF, STVREN=ON, LVP=OFF, XINST=OFF, DEBUG=OFF

// HID feature buffer
volatile unsigned char HIDFeatureBuffer[HID_FEATURE_REPORT_BYTES];


#pragma code low_vector=0x8
void low_interrupt (void)
{
}


#pragma code

// Allocate buffers in RAM for storage of bytes that have either just
// come in from the SIE or are waiting to go out to the SIE.
char txBuffer[HID_INPUT_REPORT_BYTES];
char rxBuffer[HID_OUTPUT_REPORT_BYTES];
unsigned int timestamp=0;
extern byte transferType;
#if DEBUG_PRINT

static void TxRS232(unsigned char txData)
{
    _asm clrwdt _endasm;
    _usart_putc(txData);
}

static void InitializeUSART()
{
    TRISC &= 0xBF; // Set RC6 as an output
    TRISC |= 0x80; // Set RC7 as an input
    RCSTA   = 0x90; // Enable serial port, enable receiver
    TXSTA   = 0x24; // Asynch, TSR empty, BRGH=1

    // Baud rate formula for BRG16=1, BRGH=1: Baud Rate = Fosc/(4 (n + 1)),
    // or n = (Fosc / (4 * BaudRate)) - 1
    // At 48 MHz, for 115.2K Baud:
    //     SPBRGH:SPBRG = n = Fosc / (4 * 115200) - 1 = 103.17
    BAUDCON = 0x18; // BRG16=1 txd inverted
//    SPBRGH  = 0x00; // At 48MHz, SPBRGH=0, SPBRG=103 gives 115.2K Baud
    SPBRGH  = 0x02; // 0x0270 gives 19200 Baud
#if CLK_48MHZ
//    SPBRG   = 103;  // For 48 MHz clock
    SPBRG   = 0x70;  // For 48 MHz clock
#else
    SPBRG   = 52;   // For 24 MHz clock
#endif

    printf("USB Test Startup\r\n");
}
#endif

// Regardless of what the USB is up to, we check the USART to see
// if there's something we should be doing.
#if DEBUG_PRINT
static void checkEcho()
{
	if(PIR1bits.RCIF)
    {
        // Have a character to echo
		unsigned char rxByte;
		PORTAbits.RA4 ^= 1;
		rxByte =RCREG;
		_usart_putc(rxByte);
		if (rxByte == 'X')
		{
			if(deviceState == DETACHED)
				printf("Detached\r\n");
			else if(deviceState == ATTACHED)
				printf("Attached\r\n");
			else if(deviceState == POWERED)
				printf("Powered\r\n");
			else if(deviceState == DEFAULT)
				printf("Default\r\n");
			else if(deviceState == ADDRESS)
				printf("Address\r\n");
			else if(deviceState == CONFIGURED)
				printf("Configured\r\n");
			else
				printf("Unknown state\r\n");

			printf("UCON: %x, UCFG: %x\r\n", UCON, UCFG);
        }
    }
}
#endif

// Entry point for user initialization
void UserInit(void)
{
#if DEBUG_PRINT
	InitializeUSART();
#endif
	LATA=0;
	TRISA=0b11111001;
	LATB=0;
    TRISB=0;
    T0CON=0x80;			// Timer0 period = T*2^16:2 = 5.46ms:2 = 10.93ms (PS=2)
}

#define LOBYTE(x) (*((char *)&x))
#define HIBYTE(x) (*(((char *)&x)+1))
// If we got some bytes from the host, then echo them back.
static void USBEcho(void)
{
    byte rxCnt, i;

    // Find out if an Output report has been received from the host.
    rxCnt = HIDRxReport(rxBuffer, HID_OUTPUT_REPORT_BYTES);

	// If no bytes in, then nothing to do
	if (rxCnt == 0)
		return;
	
	// put first byte in PORTB
	LATB=rxBuffer[0];
	// x01 to signal interrupt transfer
	LATBbits.LATB1=0;
	LATBbits.LATB0=1;
	
	//read PORTA levels and store in first byte of report
	txBuffer[0]=PORTA;
	txBuffer[1]=0xF0;	//this is to differentiate between different transfer types
	txBuffer[2]=HIBYTE(timestamp);
	txBuffer[3]=LOBYTE(timestamp);
    // Copy input bytes to the output buffer
	for (i=4;i<HID_OUTPUT_REPORT_BYTES;i++)
        txBuffer[i] = rxBuffer[i];            

    // As long as the SIE is owned by the processor, we let USB tasks continue.
	while (ep1Bi.Stat & UOWN)
        ProcessUSBTransactions(); 

    // The report will be sent in the next interrupt IN transfer.
    HIDTxReport(txBuffer, HID_INPUT_REPORT_BYTES);
}

// Central processing loop.  Whenever the firmware isn't busy servicing
// the USB, we will get control here to do other processing.
void ProcessIO(void)
{
#if DEBUG_PRINT
	// Process USART
    checkEcho();
#endif

    // User Application USB tasks
     if ((deviceState < CONFIGURED) || (UCONbits.SUSPND==1))
		return;

	// Process USB: Echo back any bytes that have come in.
    USBEcho();
}

// Initialization for a SET_FEATURE request.  This routine will be
// invoked during the setup stage and is used to set up the buffer
// for receiving data from the host
void SetupFeatureReport(byte reportID)
{
    if (reportID == 0)
    {
        // When the report arrives in the data stage, the data will be  
        // stored in HIDFeatureBuffer.
        inPtr = (byte*)&HIDFeatureBuffer;
    }
}

// Post processing for a SET_FEATURE request.  After all the data has
// been delivered from host to device, this will be invoked to perform
// application specific processing.
void SetFeatureReport(byte reportID)
{
#if DEBUG_PRINT
	//printf("SetFeatureReport(0x%hx)\r\n", reportID);
#endif
    // Currently only handling report 0, ignore any others.
    if (reportID == 0)
    {
        // Set the state of PORTB based on the first byte
        // of the feature report.
        LATB = HIDFeatureBuffer[0];
		// x10 to signal feature transfer
		LATBbits.LATB1=1;
		LATBbits.LATB0=0;
    }
}

// Handle a feature report request on the control pipe
void GetFeatureReport(byte reportID)
{
#if DEBUG_PRINT
	//printf("GetFeatureReport(0x%uhx): 0x%hx, 0x%hx\r\n",
	//	(byte)reportID, (byte)HIDFeatureBuffer[0],
	//	(byte)HIDFeatureBuffer[1]);
#endif

	if (reportID == 0)
	{
		// Handle report #0
		outPtr = (byte *)&HIDFeatureBuffer;
		HIDFeatureBuffer[0]=PORTA;
		HIDFeatureBuffer[1]=0xF1;	//this is to differentiate between different transfer types
		HIDFeatureBuffer[2]=HIBYTE(timestamp);
		HIDFeatureBuffer[3]=LOBYTE(timestamp);
		wCount = HID_FEATURE_REPORT_BYTES;
		transferType=0;
	}

}

// Handle control out.  This might be an alternate way of processing
// an output report, so all that's needed is to point the output
// pointer to the output buffer
// Initialization for a SET_REPORT request.  This routine will be
// invoked during the setup stage and is used to set up the buffer
// for receiving data from the host
void SetupOutputReport(byte reportID)
{
	if (reportID == 0)
	{
		// When the report arrives in the data stage, the data will be  
		// stored in HIDFeatureBuffer
		inPtr = (byte*)&HIDRxBuffer;
	}
}

// Post processing for a SET_REPORT request.  After all the data has
// been delivered from host to device, this will be invoked to perform
// application specific processing.
void SetOutputReport(byte reportID)
{
#if DEBUG_PRINT
	//printf("SetOutputReport(0x%hx)\r\n", reportID);
#endif
	// Currently only handling report 0, ignore any others.
	if (reportID != 0)
		return;

	LATB=HIDRxBuffer[0];
	// x11 to signal SET_REPORT transfer
	LATBbits.LATB1=1;
	LATBbits.LATB0=1;
}

// Handle a control input report
void GetInputReport(byte reportID)
{
#if DEBUG_PRINT
	printf("GetInputReport: 0x%uhx\r\n", reportID);
#endif
	if (reportID == 0)
	{
		byte i;
		// Send back the contents of the HID report
		// TBD: provide useful information...
		outPtr = (byte *)&HIDTxBuffer;
		HIDTxBuffer[0]=PORTA;
		HIDTxBuffer[1]=0xF2;	//this is to differentiate between different transfer types
		HIDTxBuffer[2]=HIBYTE(timestamp);
		HIDTxBuffer[3]=LOBYTE(timestamp);
		for(i=4;i<HID_INPUT_REPORT_BYTES;i++) HIDTxBuffer[i]=HIDRxBuffer[i];
		// The number of bytes in the report (from usb.h).
		wCount = HID_INPUT_REPORT_BYTES;
		transferType=0;
	}
}

// Entry point of the firmware
void main(void)
{
	byte led_cnt=0;
	// Set all I/O pins to digital
    ADCON1 |= 0x0F;
    
	// Initialize USB
    UCFG = 0x14; // Enable pullup resistors; full speed mode

    deviceState = DETACHED;
    remoteWakeup = 0x00;
    currentConfiguration = 0x00;

	// Call user initialization function
	UserInit(); 
	
#define LED1              LATAbits.LATA1
#define LED2              LATAbits.LATA2
	
	while(1)
    {
        // Ensure USB module is available
		EnableUSBModule();
		
		// As long as we aren't in test mode (UTEYE), process
		// USB transactions.
		if(UCFGbits.UTEYE != 1)
			ProcessUSBTransactions();

        // Application specific tasks
        ProcessIO();
        
        //Blink LED2 @ 1Hz when configured
	 	if(INTCONbits.TMR0IF){	
			INTCONbits.TMR0IF=0;
			timestamp++;
			led_cnt++;
			if (deviceState<CONFIGURED&&led_cnt>=10){
				LED2 = !LED2;
				led_cnt=0;
			}
			else if (deviceState == CONFIGURED&&led_cnt>=46){
				LED2 = !LED2;
				led_cnt=0;
			}
		}	
    }
}
